package com.ur.urcap.externalaxis.coordinatedmotionprogrammer.impl;

import java.net.Socket;
import java.net.ServerSocket;
import java.net.SocketTimeoutException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.DataOutputStream;

import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.data.DataModel;
import com.ur.urcap.api.domain.program.nodes.contributable.URCapProgramNode;
import com.ur.urcap.api.domain.undoredo.UndoRedoManager;
import com.ur.urcap.api.domain.undoredo.UndoableChanges;
import com.ur.urcap.api.contribution.program.ProgramAPIProvider;

// This class contains the implementation details of a generic program node that
// stores axis group, tracked frame, speed, acceleration, and blend radius. This
// contains implementation details for setting the speed, acceleration, and
// blend radius from a GUI.  It assumes that the derived types will implement
// details for the axis group and tracked frame if they need it.
public
class BaseProcessMoveNodeContribution implements BaseProcessMoveNodeInterface {
 private
  BaseProcessMoveNodeView view;

 private
  final DataModel model;

 private
  URCapProgramNode parentNode;

 protected
  final Integer SOCKET_PORT =
      12343;  // socket for getting string waypoint data back from robot

 protected
  final Integer TIMEOUT = 1000;  // time to wait for response back from robot
                                 // with waypoint data in milliseconds

 protected
  static final String AXIS_GROUP_KEY = "AXIS_GROUP";

 protected
  static final String TRACKED_FRAME_KEY = "TRACKED_FRAME";

 private
  static final String SPEED_KEY = "SPEED";

 private
  static final String ACCELERATION_KEY = "ACCEL";

 private
  static final String BLEND_RADIUS_KEY = "RADIUS";

 private
  static final Double DEFAULT_SPEED =
      0.25;  // m/s (from URScript manual for movep and movec)

 private
  static final Double DEFAULT_ACCELERATION =
      1.2;  // m/s/s (from URScript manual for movep and movec)

 private
  static final Double DEFAULT_BLEND_RADIUS =
      0.0;  // m (from URScript manual for movep and movec)

 protected
  final UndoRedoManager undoRedoManager;

 public
  BaseProcessMoveNodeContribution(ProgramAPIProvider apiProvider,
                                  DataModel model,
                                  BaseProcessMoveNodeView view) {
    this.model = model;
    this.view = view;
    this.undoRedoManager = apiProvider.getProgramAPI().getUndoRedoManager();
  }

 public
  void openView() {
    view.setSpeedInputField(getSpeed());
    view.setAccelerationInputField(getAcceleration());
    view.setBlendRadiusInputField(getBlendRadius());
  }

 public
  String getAxisGroup() { return model.get(AXIS_GROUP_KEY, ""); }

 public
  String getTrackedFrame() { return model.get(TRACKED_FRAME_KEY, ""); }

 public
  Double getSpeed() { return model.get(SPEED_KEY, DEFAULT_SPEED); }

 public
  Double getAcceleration() {
    return model.get(ACCELERATION_KEY, DEFAULT_ACCELERATION);
  }

 public
  Double getBlendRadius() {
    return model.get(BLEND_RADIUS_KEY, DEFAULT_BLEND_RADIUS);
  }

 public
  void setAxisGroup(final String value) { model.set(AXIS_GROUP_KEY, value); }

 public
  void setTrackedFrame(final String value) {
    model.set(TRACKED_FRAME_KEY, value);
  }

 public
  void setSpeed(final Double value) {
    view.setSpeedInputField(value);
    model.set(SPEED_KEY, value);
  }

 public
  void setAcceleration(final Double value) {
    view.setAccelerationInputField(value);
    model.set(ACCELERATION_KEY, value);
  }

 public
  void setBlendRadius(final Double value) {
    view.setBlendRadiusInputField(value);
    model.set(BLEND_RADIUS_KEY, value);
  }

 public
  KeyboardInputCallback<Double> getCallbackForSpeed() {
    return new KeyboardInputCallback<Double>() {
      @Override public void onOk(final Double value) {
        setSpeed(value);
      }
    };
  }

 public
  KeyboardInputCallback<Double> getCallbackForAcceleration() {
    return new KeyboardInputCallback<Double>() {
      @Override public void onOk(Double value) {
        setAcceleration(value);
      }
    };
  }

 public
  KeyboardInputCallback<Double> getCallbackForBlendRadius() {
    return new KeyboardInputCallback<Double>() {
      @Override public void onOk(Double value) {
        setBlendRadius(value);
      }
    };
  }

  // This runs inside a thread to receive a string over a socket from a primary
  // program running on the robot.
 private
  class StringFromRobotGetter implements Runnable {
   private
    String receivedString = "";

    @Override public void run() {
      try {
        ServerSocket server = new ServerSocket(SOCKET_PORT);

        // Set a timeout on the server socket so that it's cleaned up nicely
        // if an error occurs on the program running on the robot and
        // doesn't close the socket connection itself.
        server.setSoTimeout(TIMEOUT);

        Socket socket;
        try {
          socket = server.accept();
        } catch (SocketTimeoutException e) {
          server.close();
          throw e;
        }

        BufferedReader reader =
            new BufferedReader(new InputStreamReader(socket.getInputStream()));

        receivedString = reader.readLine();

        reader.close();
        socket.close();
        server.close();
      } catch (IOException e) {
        System.err.println("Failed to get the waypoint data from the robot.");
        receivedString = "";
      }
    }

   public
    String getReceivedString() { return receivedString; }
  }

  // Sends a program to the controller over the primary interface.
  private void
  sendPrimaryProgramToRobot(final String primaryProgram) {
    try {
      Socket socket = new Socket("127.0.0.1",
                                 30001);  // primary interface is port 30001

      DataOutputStream outputStream =
          new DataOutputStream(socket.getOutputStream());
      outputStream.write(primaryProgram.getBytes("US-ASCII"));
      outputStream.flush();

      socket.close();
    } catch (IOException e) {
      System.err.println("Failed to write the primary program to the socket.");
    }
  }

  // Retrieves a string from a program running on the robot. The program
  // running on the robot should send the string via socket to this Java node.
 public
  String getStringResponseBackFromRobot(final String programToRunOnRobot) {
    StringFromRobotGetter myStringGetter = new StringFromRobotGetter();
    Thread myThread = new Thread(myStringGetter);

    myThread.start();

    sendPrimaryProgramToRobot(programToRunOnRobot);

    try {
      myThread.join();
    } catch (InterruptedException e) {
    }

    return myStringGetter.getReceivedString();
  }
}
