package com.ur.urcap.externalaxis.coordinatedmotionprogrammer.impl;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.StringReader;
import java.net.Socket;
import java.nio.file.Files;
import java.nio.file.Paths;

import com.ur.urcap.api.contribution.URCapProgramNodeService;
import com.ur.urcap.api.contribution.ProgramNodeContribution;
import com.ur.urcap.api.contribution.program.ProgramAPIProvider;
import com.ur.urcap.api.domain.data.DataModel;
import com.ur.urcap.api.domain.script.ScriptWriter;
import com.ur.urcap.api.domain.undoredo.UndoRedoManager;
import com.ur.urcap.api.domain.undoredo.UndoableChanges;
import com.ur.urcap.api.domain.ProgramAPI;
import com.ur.urcap.api.domain.program.ProgramModel;
import com.ur.urcap.api.domain.program.nodes.ProgramNodeFactory;
import com.ur.urcap.api.domain.program.structure.TreeNode;
import com.ur.urcap.api.domain.program.structure.TreeStructureException;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.program.nodes.ProgramNode;
import com.ur.urcap.api.domain.program.nodes.contributable.URCapProgramNode;
import com.ur.urcap.api.domain.program.structure.ProgramNodeVisitor;

// This implements a Polyscope node that
public
class CoordinatedProcessProgramNodeContribution extends
    BaseProcessMoveNodeContribution implements ProgramNodeContribution {
 private
  final CoordinatedProcessProgramNodeView view;

 private
  final DataModel model;

 private
  final UndoRedoManager undoRedoManager;

 private
  final ProgramAPIProvider apiProvider;

 public
  CoordinatedProcessProgramNodeContribution(
      ProgramAPIProvider apiProvider,
      CoordinatedProcessProgramNodeView view,
      DataModel model) {
    super(apiProvider, model, view);

    this.view = view;
    this.model = model;
    this.undoRedoManager = apiProvider.getProgramAPI().getUndoRedoManager();
    this.apiProvider = apiProvider;
  }

  // Adds a MovePWithAxisGroup or MoveCWithAxisGroup to this node's subtree.
 private void addMoveWithAxisGroup(Class<? extends URCapProgramNodeService> newNodeType) {
    ProgramNodeFactory nodeFactory =
        apiProvider.getProgramAPI().getProgramModel().getProgramNodeFactory();
    TreeNode root =
        apiProvider.getProgramAPI().getProgramModel().getRootTreeNode(this);

    apiProvider.getProgramAPI().getUndoRedoManager().recordChanges(
        new UndoableChanges() {
          @Override public void executeChanges() {
            try {
              URCapProgramNode newChild =
                  (URCapProgramNode)root
                      .addChild(nodeFactory.createURCapProgramNode(newNodeType))
                      .getProgramNode();

              BaseProcessMoveNodeInterface newChildNodeInterface =
                  newChild.getAs(BaseProcessMoveNodeInterface.class);

              newChildNodeInterface.setAxisGroup(getAxisGroup());
              newChildNodeInterface.setTrackedFrame(getTrackedFrame());
              newChildNodeInterface.setSpeed(getSpeed());
              newChildNodeInterface.setAcceleration(getAcceleration());
              newChildNodeInterface.setBlendRadius(getBlendRadius());
            } catch (TreeStructureException e) {
            }
          }
        });
  }

  // Adds a MovePWithAxisGroup to the program tree when a button is pressed on
  // the UI.
 public
  void addMovePWithAxisGroup() {
    addMoveWithAxisGroup(MovePWithAxisGroupNodeService.class);
  }

  // Adds a MoveCWithAxisGroup to the program tree when a button is pressed on
  // the UI.
 public
  void addMoveCWithAxisGroup() {
    addMoveWithAxisGroup(MoveCWithAxisGroupNodeService.class);
  }

  @Override public void openView() {
    super.openView();

    if (getAxisGroup() != "") {
      view.setAxisGroupComboBox(getAxisGroup());
    }

    if (getTrackedFrame() != "") {
      view.setTrackedFrameComboBox(getTrackedFrame());
    }
  }

 public
  String getAxisGroup() { return model.get(AXIS_GROUP_KEY, ""); }

 public
  String getTrackedFrame() { return model.get(TRACKED_FRAME_KEY, ""); }

  @Override public void closeView() {
  }

  @Override public String getTitle() {
    return "Coordinated Move";
  }

  @Override public boolean isDefined() {
    return ((getAxisGroup() != "") && (getTrackedFrame() != ""));
  }

  // Writes URScript that adds a frame to the world model that is placed
  // relative to the tracked frame and is attached to a parent. This is used
  // for visualization and debugging only. It's not needed for commanding the
  // robot!
 private
  void addFrameToWorldModel(ScriptWriter writer,
                            final String childName,
                            final String parentName,
                            final String poseInTrackedFrame) {
    writer.appendLine(String.format("if (is_frame(\"%s\")):", childName));
    writer.appendLine(String.format("  move_frame(\"%s\", %s, \"%s\")",
                                    childName, poseInTrackedFrame,
                                    getTrackedFrame()));
    writer.appendLine(String.format("else:"));
    writer.appendLine(String.format("  add_frame(\"%s\", %s, \"%s\")",
                                    childName, poseInTrackedFrame,
                                    getTrackedFrame()));
    writer.appendLine(String.format("end"));
    writer.appendLine(
        String.format("attach_frame(\"%s\", \"%s\")", childName, parentName));
  }

  // Writes URScript that adds frames for all the movep and movec waypoints.
  // This is used for visualization and debugging only. It's not needed for
  // commanding the robot!
  void addWorldModelTrajectoryToURScript(ScriptWriter writer) {
    final String NODE_WAYPOINT_FRAME_PREFIX =
        "nodeWaypointFrame";  // frames placed at the waypoint targets have this
    // prefix followed by an integer

    final String NODE_VIAPOINT_FRAME_PREFIX =
        "nodeViaFrame";  // frames placed at the viapoint of movec will have
                         // this prefix followed by an integer

    final TreeNode root =
        apiProvider.getProgramAPI().getProgramModel().getRootTreeNode(this);

    // This writes URScript to place world model frames at each of the waypoints
    // for later visualization in RViz. Each waypoint from is attached to the
    // prior one.
    root.traverse(new ProgramNodeVisitor() {
     private
      Integer frameIndex =
          0;  // used to make goal frames name unique in the world model

      @Override public void visit(final URCapProgramNode programNode, int index,
                                  int depth) {
        // Add a single frame for the movep node
        if (programNode.canGetAs(MovePWithAxisGroupNodeInterface.class)) {
          MovePWithAxisGroupNodeInterface movep =
              programNode.getAs(MovePWithAxisGroupNodeInterface.class);

          String robotWaypoint = movep.getRobotWaypoint();
          String waypointFrameToAdd =
              NODE_WAYPOINT_FRAME_PREFIX + frameIndex.toString();
          String frameToAttachWaypointFrameTo = getTrackedFrame();

          if (frameIndex > 0) {
            Integer previousNodeIndex = frameIndex - 1;
            frameToAttachWaypointFrameTo =
                NODE_WAYPOINT_FRAME_PREFIX + previousNodeIndex.toString();
          }

          addFrameToWorldModel(writer, waypointFrameToAdd,
                               frameToAttachWaypointFrameTo, robotWaypoint);

          frameIndex += 1;
        }

        // Add a frame for the movec end and one for the via
        else if (programNode.canGetAs(MoveCWithAxisGroupNodeInterface.class)) {
          MoveCWithAxisGroupNodeInterface movec =
              programNode.getAs(MoveCWithAxisGroupNodeInterface.class);

          // Add the via frame to the world model and attach it to the previous
          // waypoint

          String robotViaPoint = movec.getRobotViaPoint();
          String viaFrameToAdd =
              NODE_VIAPOINT_FRAME_PREFIX + frameIndex.toString();

          // Attach the via frame to the tracked frame or the last waypoint
          // frame if there was one.
          String frameToAttachViaFrameTo = getTrackedFrame();
          if (frameIndex > 0) {
            Integer previousFrameIndex = frameIndex - 1;
            frameToAttachViaFrameTo =
                NODE_WAYPOINT_FRAME_PREFIX + previousFrameIndex.toString();
          }

          addFrameToWorldModel(writer, viaFrameToAdd, frameToAttachViaFrameTo,
                               robotViaPoint);

          frameIndex += 1;

          // Add the waypoint frame to the world model and attach it to the via
          // point

          String robotWaypoint = movec.getRobotWaypoint();
          String waypointFrameToAdd =
              NODE_WAYPOINT_FRAME_PREFIX + frameIndex.toString();

          String frameToAttachWaypointFrameTo = getTrackedFrame();

          Integer previousNodeIndex = frameIndex - 1;
          frameToAttachWaypointFrameTo =
              NODE_VIAPOINT_FRAME_PREFIX + previousNodeIndex.toString();

          addFrameToWorldModel(writer, waypointFrameToAdd,
                               frameToAttachWaypointFrameTo, robotWaypoint);

          frameIndex += 1;
        }
      }
    });
  }

  // This writes the URScript code that makes the robot follow the programmed
  // movep/movec trajectory.
 private
  void addProcessTrajectoryToURScript(ScriptWriter writer) {
    TreeNode root =
        apiProvider.getProgramAPI().getProgramModel().getRootTreeNode(this);

    writer.appendLine(
        String.format("frame_tracking_enable(\"%s\")", getTrackedFrame()));

    // This writes the URScript command for each node.
    root.traverse(new ProgramNodeVisitor() {
      @Override public void visit(URCapProgramNode programNode, int index,
                                  int depth) {
        // Add a movep_with_axis_group command.
        if (programNode.canGetAs(MovePWithAxisGroupNodeInterface.class)) {
          final MovePWithAxisGroupNodeInterface movep =
              programNode.getAs(MovePWithAxisGroupNodeInterface.class);

          final String axisWaypoint = movep.getAxisWaypoint();
          final String robotWaypoint = movep.getRobotWaypoint();
          final String groupName = movep.getAxisGroup();
          final Double speed = movep.getSpeed();
          final Double acceleration = movep.getAcceleration();
          final Double blendRadius = movep.getBlendRadius();

          writer.appendLine(String.format(
              "movep_with_axis_group(convert_pose(%s, \"%s\", " +
                  "\"base\"), a=%s, v=%s, r=%s, name=\"%s\", axis_target=%s)",
              robotWaypoint, getTrackedFrame(), acceleration.toString(),
              speed.toString(), blendRadius.toString(), groupName,
              axisWaypoint));
        }

        // Add a movec_with_axis_group command.
        else if (programNode.canGetAs(MoveCWithAxisGroupNodeInterface.class)) {
          final MoveCWithAxisGroupNodeInterface movec =
              programNode.getAs(MoveCWithAxisGroupNodeInterface.class);

          final String axisWaypoint = movec.getAxisWaypoint();
          final String robotWaypoint = movec.getRobotWaypoint();
          final String robotViaPoint = movec.getRobotViaPoint();
          final String groupName = movec.getAxisGroup();
          final Double speed = movec.getSpeed();
          final Double acceleration = movec.getAcceleration();
          final Double blendRadius = movec.getBlendRadius();

          writer.appendLine(String.format(
              "movec_with_axis_group(convert_pose(%s, \"%s\", \"base\"), " +
                  "convert_pose(%s, \"%s\", \"base\"), a=%s, v=%s, r=%s, " +
                  "name=\"%s\", axis_target=%s)",
              robotViaPoint, getTrackedFrame(), robotWaypoint,
              getTrackedFrame(), acceleration.toString(), speed.toString(),
              blendRadius.toString(), groupName, axisWaypoint));
        }
      }
    });

    writer.appendLine(String.format("frame_tracking_disable()"));
  }

  @Override public void generateScript(ScriptWriter writer) {
    addWorldModelTrajectoryToURScript(
        writer);  // for debugging and visualization only!
    addProcessTrajectoryToURScript(
        writer);  // makes the robot follow the programmed movec/movep
                  // trajectory
  }

 public
  void saveTrackedFrame(final String value) {
    undoRedoManager.recordChanges(new UndoableChanges() {
      @Override public void executeChanges() {
        model.set(TRACKED_FRAME_KEY, value);
      }
    });
  }

 public
  void saveAxisGroup(final String value) {
    undoRedoManager.recordChanges(new UndoableChanges() {
      @Override public void executeChanges() {
        model.set(AXIS_GROUP_KEY, value);
      }
    });
  }

 public
  String getAxisAndFrameStringProgram() {
    // Construct the primary program to be sent to the controller
    StringBuilder programBuilder = new StringBuilder();

    programBuilder.append("sec mySecondaryProgram():\n");

    programBuilder.append("  def get_axes_recursively(node, stringAxes):\n");
    programBuilder.append("    axes = get_children(node)\n");

    programBuilder.append("    index = 0\n");
    programBuilder.append("    while index < length(axes):\n");
    programBuilder.append("      axis = axes[index]\n");
    programBuilder.append("      if axis !=\"flange\":\n");
    programBuilder.append("        stringAxes = stringAxes + \",\"\n");
    programBuilder.append("        stringAxes = stringAxes + axis\n");
    programBuilder.append(
        "        stringAxes = get_axes_recursively(axis, stringAxes)\n");
    programBuilder.append("      end\n");
    programBuilder.append("      index = index + 1\n");
    programBuilder.append("    end\n");
    programBuilder.append("    return stringAxes\n");
    programBuilder.append("  end\n");

    programBuilder.append(
        "  str = get_axes_recursively(\"world\", \"world\")\n");

    programBuilder.append("  socket_open(\"127.0.0.1\", " +
                          SOCKET_PORT.toString() + ", \"my_socket\")\n");
    programBuilder.append("  socket_send_line( str, \"my_socket\")\n");
    programBuilder.append("  socket_close(\"my_socket\")\n");

    programBuilder.append("end\n");  // the unindented "end" marks the end of
                                     // the primary program

    return programBuilder.toString();
  }

 public
  String getAxisGroupsStringProgram() {
    // Construct the primary program to be sent to the controller
    StringBuilder programBuilder = new StringBuilder();

    programBuilder.append("sec mySecondaryProgram():\n");

    programBuilder.append("  def get_axis_groups():\n");
    programBuilder.append("    stringGroups = \"\"");
    programBuilder.append("    child_nodes = get_children(\"world\")");
    programBuilder.append("    index = 0");
    programBuilder.append("    while index < length(child_nodes):");
    programBuilder.append("      child_node = child_nodes[index]");
    programBuilder.append("      if not is_frame(child_node):");
    programBuilder.append("        if (stringGroups != \"\"):");
    programBuilder.append("          stringGroups = stringGroups + \",\"");
    programBuilder.append("        end");
    programBuilder.append("        stringGroups = stringGroups + child_node");
    programBuilder.append("      end");
    programBuilder.append("      index = index + 1");
    programBuilder.append("    end");
    programBuilder.append("    return stringGroups");
    programBuilder.append("  end");

    programBuilder.append("  str = get_axis_groups()\n");

    programBuilder.append("  socket_open(\"127.0.0.1\", " +
                          SOCKET_PORT.toString() + ", \"my_socket\")\n");
    programBuilder.append("  socket_send_line( str, \"my_socket\")\n");
    programBuilder.append("  socket_close(\"my_socket\")\n");

    programBuilder.append("end\n");  // the unindented "end" marks the end of
                                     // the primary program

    return programBuilder.toString();
  }

 public
  String[] getAxisAndFrameNames() {
    final String worldModelString =
        getStringResponseBackFromRobot(getAxisAndFrameStringProgram());
    return worldModelString.split(",");
  }

 public
  String[] getAxisGroupNames() {
    final String axisGroupsString =
        getStringResponseBackFromRobot(getAxisGroupsStringProgram());
    return axisGroupsString.split(",");
  }
}
