package com.ur.urcap.externalaxis.moveto.toolbar;

import com.ur.urcap.api.contribution.toolbar.ToolbarAPIProvider;
import com.ur.urcap.api.contribution.toolbar.ToolbarContext;
import com.ur.urcap.api.contribution.toolbar.swing.SwingToolbarContribution;
import com.ur.urcap.api.domain.userinteraction.inputvalidation.InputValidationFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardTextInput;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptCommand;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptReader;
import com.ur.urcap.externalaxis.moveto.impl.Style;
import com.ur.urcap.externalaxis.moveto.impl.V5Style;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEClient;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEOutputHandler;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEOutputParam;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptSender;

import javax.swing.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.PopupMenuListener;
import javax.swing.event.PopupMenuEvent;
import javax.swing.GroupLayout;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.stream.Collectors;

class MoveToToolbarContribution implements SwingToolbarContribution {
	private final Style style;
	private final ToolbarAPIProvider apiProvider;
	private final InputValidationFactory validatorFactory;
	private final KeyboardInputFactory keyboardInputFactory;
	private static final int MAXIMUM_FRACTION_DIGITS = 3;
	private static final Integer SOCKET_PORT = 12343;

	// ScriptCommunicator
	private final ScriptSender sender;

	// RTDE
	private final RTDEClient client;

	// DecimalFormat
	private DecimalFormat df = new DecimalFormat("#");
	
	// GUI
	private final JLabel axisGroupLabel;
	private final JComboBox axisGroupInput;
	private final JLabel errorDisplay;
	private final JLabel targetPositionsLabel;
	private final JTextField targetPositionsInput;
	private final JLabel actualPositionsLabel;
	private final JLabel actualPositionsDisplay;
	private final JLabel frameTrackingLabel;
	private final JCheckBox frameTrackingCheckBox;
	private final JTextField frameName;
	private final JButton moveToButton;
	private final JSlider speedSlider;
	private final JLabel speedSliderLabel;
	private final JSeparator separator;
	private final JLabel sentCommandLabel;
	private final JTextArea sentCommandArea;

	// Helper Variables
	private Integer[] groupIndices = new Integer[0];
	private String receivedString = "";

	MoveToToolbarContribution(ToolbarContext context) {
		// API and Keyboard
		apiProvider = context.getAPIProvider();
		style = new V5Style();
		keyboardInputFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getKeyboardInputFactory();
		validatorFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getInputValidationFactory();

		// GUI
		axisGroupLabel = new JLabel("Axis Group:");
		axisGroupInput = new JComboBox();
		errorDisplay = new JLabel();
		targetPositionsLabel = new JLabel("Target Positions:");
		targetPositionsInput = new JTextField("[]");
		actualPositionsLabel = new JLabel("Actual Positions:");
		actualPositionsDisplay = new JLabel();
		frameTrackingLabel = new JLabel("Frame Tracking:");
		frameTrackingCheckBox = new JCheckBox("enable");
		frameName = new JTextField();
		moveToButton = new JButton("Move to Target Positions");
		speedSlider = new JSlider();
		speedSliderLabel = new JLabel("Speed [%]");
		separator = new JSeparator();
		sentCommandLabel = new JLabel("Sent Command:");
		sentCommandArea = new JTextArea();

		sender = new ScriptSender();
		client = new RTDEClient(new RTDEOutputHandler() {
			@Override
			public void onProtocolVersion(int protocolVersion) {
				if (protocolVersion != 2) {
					throw new RuntimeException("Protocol v2 expected");
				}
			}

			@Override
			public void onData(RTDEOutputParam output, Object value) {
					double[] data = (double[]) value;
					if (groupIndices.length > 0) {
						String stringData = Arrays.stream(groupIndices)
							.map(i -> df.format(data[i]))
							.collect(Collectors.joining(", "));
						actualPositionsDisplay.setText("[" + stringData + "]");
					}
					else {
						actualPositionsDisplay.setText("[]");
					}
				}
			});
		df.setMaximumFractionDigits(MAXIMUM_FRACTION_DIGITS);
	}

	@Override
	public void openView() {
		// Connect to the ScriptSender and RTDEClient
		sender.connect("127.0.0.1", ScriptSender.PRIMARY_PORT, 200);
		if (client.connect("127.0.0.1", RTDEClient.PORT, 200)) {
			try {
				client.requestProtocolVersion(2);
				client.setupOutputsV2(50, RTDEOutputParam.actual_external_q);
				client.startOutput();
			} catch (Exception e) {
				System.out.println("RTDE client setup failed");
				System.out.println(e);
			}
		}
	}

	@Override
	public void closeView() {
		sender.disconnect();
		client.disconnect();
	}

	public void buildUI(JPanel jPanel) {
		setupComboBox(axisGroupInput, new PopupMenuListener() {
			@Override
			public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
				axisGroupInput.removeAllItems();
				String[] names = getAxisGroupNames();
				for (String name : names) {
					axisGroupInput.addItem(name);
				}
			}

			@Override
			public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
				String[] axisIndices = getGroupIndices(axisGroupInput.getSelectedItem().toString());
				System.out.println(axisIndices);
				if (axisIndices.length > 0) {
					groupIndices = Arrays.stream(axisIndices).map(Integer::parseInt).toArray(Integer[]::new);
					errorDisplay.setText("");
				}
				else {
					groupIndices = new Integer[0];
					errorDisplay.setText("No Axes Defined");
					errorDisplay.setForeground(Color.RED);
				}
			}

			@Override
			public void popupMenuCanceled(PopupMenuEvent e) {
			}
		});

		setupTextInputField(targetPositionsInput, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForAxisGroupPosition();
				keyboard.show(targetPositionsInput, getCallbackForAxisGroupPosition());
			}
		});

		setupDisplayField(actualPositionsDisplay);

		frameTrackingCheckBox.setEnabled(false);
		frameTrackingCheckBox.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (!frameTrackingCheckBox.isSelected()) {
					frameName.setText("");
				}
			}
		});
		setupTextInputField(frameName, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForTrackedFrameName();
				keyboard.show(frameName, getCallbackForTrackedFrameName());
			}
		});

		moveToButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				moveAxisButtonMousePressed();
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				moveAxisButtonMouseReleased();
			}
		});
		moveToButton.setPreferredSize(style.getButtonSize());
		moveToButton.setMaximumSize(moveToButton.getPreferredSize());
		moveToButton.setMinimumSize(moveToButton.getPreferredSize());

		ChangeListener changeListener = e -> {
			speedSliderLabel.setText("Speed " + speedSlider.getValue() + "%");
		};
		speedSlider.addChangeListener(changeListener);
		speedSlider.setMinimum(1);

		sentCommandArea.setPreferredSize(style.getCommandfieldSize());
		sentCommandArea.setMaximumSize(sentCommandArea.getPreferredSize());

		GroupLayout layout = new GroupLayout(jPanel);
		jPanel.setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);
		layout.setHorizontalGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
					.addComponent(axisGroupLabel)
					.addComponent(targetPositionsLabel)
					.addComponent(actualPositionsLabel)
					.addComponent(frameTrackingLabel)
					.addComponent(sentCommandLabel)
				)
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
					.addGroup(layout.createSequentialGroup()
						.addComponent(axisGroupInput)
						.addComponent(errorDisplay)
					)
					.addComponent(targetPositionsInput)
					.addComponent(actualPositionsDisplay)
					.addGroup(layout.createSequentialGroup()
						.addComponent(frameName)
						.addComponent(frameTrackingCheckBox)
					)
				)
			)
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.CENTER)
				.addComponent(moveToButton)
				.addGroup(layout.createSequentialGroup()
					.addComponent(speedSlider)
					.addComponent(speedSliderLabel)
				)
			)
			.addComponent(separator)
			.addComponent(sentCommandArea)
		);

		layout.setVerticalGroup(layout.createSequentialGroup()
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
				.addComponent(axisGroupLabel)
				.addComponent(axisGroupInput)
				.addComponent(errorDisplay)
			)
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
				.addComponent(targetPositionsLabel)
				.addComponent(targetPositionsInput)
			)
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
				.addComponent(actualPositionsLabel)
				.addComponent(actualPositionsDisplay)
			)
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
				.addComponent(frameTrackingLabel)
				.addComponent(frameTrackingCheckBox)
				.addComponent(frameName)
			)
			.addGap(10, 10, 10)
			.addComponent(moveToButton)
			.addGap(10, 10, 10)
			.addGroup(layout.createParallelGroup(GroupLayout.Alignment.CENTER)
				.addComponent(speedSlider)
				.addComponent(speedSliderLabel)
			)
			.addGap(10, 10, 10)
			.addComponent(separator)
			.addGap(10, 10, 10)
			.addComponent(sentCommandLabel)
			.addComponent(sentCommandArea)
		);
	}

	private void setupComboBox(final JComboBox box, PopupMenuListener menuListener) {
		box.setFocusable(false);
		box.setPreferredSize(style.getInputfieldSize());
		box.setMaximumSize(box.getPreferredSize());
		box.addPopupMenuListener(menuListener);
	}

	private String[] getAxisGroupNames() {
		ScriptCommand cmd = new ScriptCommand("get_group_names");
		cmd.appendLine("retString = \"\"");
		cmd.appendLine("groups = get_children(\"world\")");
		cmd.appendLine("while groups.length() > 0:");
		cmd.appendLine("  group = groups.remove(0)");
		cmd.appendLine("  if not is_frame(group):");
		cmd.appendLine("    retString = retString + group + \",\"");
		cmd.appendLine("  end");
		cmd.appendLine("end");
		cmd.appendLine("retString = retString + \",\""); // prevent empty string
		cmd.appendLine("socket_open(\"127.0.0.1\", " + SOCKET_PORT.toString() + ", \"my_socket\")");
		cmd.appendLine("socket_send_line(retString, \"my_socket\")");
		cmd.appendLine("socket_close(\"my_socket\")");
		sendCommandAndReceiveResponse(cmd);
		return receivedString.split(",");
	}

	private String[] getGroupIndices(String groupName) {
		ScriptCommand cmd = new ScriptCommand("get_group_indices");
		// get axis groupIndices of the given group
		cmd.appendLine("axes = make_list(6,0,6)");
		cmd.appendLine("def get_axes_recursively(node):");
		cmd.appendLine("  chs = get_children(node)");
		cmd.appendLine("  while chs.length() > 0:");
		cmd.appendLine("    ch = chs.remove(0)");
		cmd.appendLine("    if not is_frame(ch):");
		cmd.appendLine("      axes[axis_group_get_axis_index(ch)] = 1");
		cmd.appendLine("      get_axes_recursively(ch)");
		cmd.appendLine("    end");
		cmd.appendLine("  end");
		cmd.appendLine("end");
		cmd.appendLine("get_axes_recursively(\"" + groupName + "\")");
		// prepare return string
		cmd.appendLine("retString = \"\"");
		cmd.appendLine("idx = 0");
		cmd.appendLine("while idx < axes.length():");
		cmd.appendLine("  if axes[idx] == 1:");
		cmd.appendLine("    if length(retString) == 0:");
		cmd.appendLine("      retString = to_str(idx)");
		cmd.appendLine("    else:");
		cmd.appendLine("      retString = retString + \",\" + to_str(idx)");
		cmd.appendLine("    end");
		cmd.appendLine("  end");
		cmd.appendLine("  idx = idx + 1");
		cmd.appendLine("end");
		cmd.appendLine("retString = retString + \",\""); // prevent empty string
		cmd.appendLine("socket_open(\"127.0.0.1\", " + SOCKET_PORT.toString() + ", \"my_socket\")");
		cmd.appendLine("socket_send_line(retString, \"my_socket\")");
		cmd.appendLine("socket_close(\"my_socket\")");
		sendCommandAndReceiveResponse(cmd);
		return receivedString.split(",");
	}

	private boolean sendCommandAndReceiveResponse(ScriptCommand command) {
		// Send the command to the robot and read the result by starting
		// a server socket and reading the data sent by the robot.
		ScriptReader reader = new ScriptReader();
		Thread thread = new Thread(reader);
		thread.start();

		if (sender.executeURScript(command)){
			try {
				thread.join();
			} catch (InterruptedException e) {
			}
			receivedString = reader.getReceivedString();
			return true;
		} else {
			return false;
		}
	}

	private void setupTextInputField(final JTextField inputField, MouseAdapter mouseAdapter) {
		inputField.setFocusable(false);
		inputField.setPreferredSize(style.getInputfieldSize());
		inputField.setMaximumSize(inputField.getPreferredSize());
		inputField.addMouseListener(mouseAdapter);
	}

	private KeyboardTextInput getKeyboardForAxisGroupPosition() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(targetPositionsInput.getText());
		return keyboardInput;
	}

	private KeyboardInputCallback getCallbackForAxisGroupPosition() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				targetPositionsInput.setText(value);
			}
		};
	}

	// Constructs the primary program to be sent over a socket. The builtin URScript command
	// 'axis_group_movej' is used to move the axis group to the desired position.
	// If the frame tracking checkbox is selected, the robot will track the given frame or axis
	// which was input by the user.
	private void moveAxisButtonMousePressed() {
		ScriptCommand cmd = new ScriptCommand("move_to");
		if (frameTrackingCheckBox.isSelected()) {
			cmd.appendLine(String.format("frame_tracking_enable(\"%s\")", frameName.getText()));
		}
		cmd.appendLine(String.format(
			"axis_group_movej(\"%s\", %s, 0.5, %s)",
			axisGroupInput.getSelectedItem().toString(),
			targetPositionsInput.getText(),
			speedSlider.getValue() / 100f)
		);
		sentCommandArea.setText(cmd.content());
		sender.executeURScript(cmd);
	}

	// Stops the movement of the axis group by calling the builtin URScript command 'axis_group_speedj',
	// and setting the speed to "0.0" for each of the axes in the group.
	private void moveAxisButtonMouseReleased() {
		double[] zeros = new double[groupIndices.length];
		ScriptCommand cmd = new ScriptCommand("stop");
		if (frameTrackingCheckBox.isSelected()){
			cmd.appendLine(String.format("frame_tracking_enable(\"%s\")", frameName.getText()));
		};
		cmd.appendLine(String.format("axis_group_speedj(\"%s\", [%s], 1)",
			axisGroupInput.getSelectedItem().toString(),
			Arrays.stream(zeros).mapToObj(df::format).collect(Collectors.joining(", ")))
		);
		if (frameTrackingCheckBox.isSelected()){
			cmd.appendLine("frame_tracking_disable()");
		};
		sentCommandArea.setText(cmd.content());
		sender.executeURScript(cmd);
	}

	private KeyboardTextInput getKeyboardForTrackedFrameName() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(frameName.getText());
		keyboardInput.setErrorValidator(validatorFactory.createStringLengthValidator(1, 255));
		return keyboardInput;
	}

	private KeyboardInputCallback<String> getCallbackForTrackedFrameName() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				frameName.setText(value);
				frameTrackingCheckBox.setEnabled(true);
			}
		};
	}

	private void setupDisplayField(final JLabel displayLabel) {
		displayLabel.setFocusable(false);
		displayLabel.setPreferredSize(style.getDisplayfieldSize());
		displayLabel.setMaximumSize(displayLabel.getPreferredSize());
	}
}
